=head1 NAME
	
	Spamalyzer::Letter

=head1 SYNOPSIS

	Letter will extract various interesting fields from both the headers
	and body of an e-mail. It's intended use is to analyze spam for
	statistics, fingerprints, etc.

	In particular it will extract :
	From the headers :
		 From:
		 Date:
		 Subject:
		 Message-ID:
		 Date received: (This is the ISP date stamp, more reliable than Date:)
		 Last relay point (what machine handed it off to your ISP?)
		 Last relay ip (IP address of the machine above)
		 Previous relay point (what machine handed it off to the last relay?)
		 Previous relay ip (IP address of the machine above)
	 From the body :
		  US Telephone numbers
		  URL's
		  e-mail addresses
		  US snail mail addresses

=head1 NOTES

	Snail mail addresses and telephone numbers are for US only. Feel free
	to extend this to international if you wish. 

	For snail mail addresses, I'll substitute the state postal code if the
	original has the state name spelled out.

	Phone numbers are canonicalized to nnn-nnn-nnnn with a possible 1- in
	front. I don't trap extensions.

	I only get 5 digit zip. I could do 9 digits, but it is pretty rarely used.

	Philosophically, I'm willing to grab a few extra things that I have
	to delete by hand, so if I have a bias, it is to have false positives.
	Of course, I haven't measured the false negatives.

=head1 EXAMPLE

	require Spamalyzer::Letter;
	my $type = 'aol';
	my $dir = '/mnt/d_drive/alan/desk/spam/keepspam';
	#  key for finding last relay header line, isp dependent
	#my $isp = 'by icct.net';
	my $isp = 'by \S+\aol.com';
	my ($file, $letter, $i);

	opendir DIR, $dir or die "$0: failed opendir [$dir]: $!\n";
	 ARTICLE:
	  while (defined($file = readdir(DIR))) {
			$i++;
			print STDERR "\rProcessed $i letters "
			my $fullfile = $dir . '/' . $file;
			$letter = new Spamalyzer::Letter $fullfile, $type, $isp;
	  		my $from = $letter->from;
			my $date = $letter->date;
			my $subject = $letter->subject;
			my $message_id = $letter->message_id;
			my $relay = $letter->lastrelay;
			my $relayip = $letter->lastrelayip;
			my $date_recvd = $letter->date_recvd;
			my $prevrelay = $letter->prevrelay;
			my $prevrelayip = $letter->prevrelayip;
			print "$file : $from : $date : $subject : $message_id : $relay : $relayip : $prevrelay : $prevrelayip : $date_recvd \n";

			#  now to the body info. The @contacts array will let you know
			#	*if* a phone, url, email, or address was found.
			my @contacts = @{[$letter->contacts]};
			my (@phone, @addr, @email, @url);
			# since there may be multiple values per letter, an array
			# is returned.

			foreach (@contacts) {
				last if /NONE/;
				@phone = @{[$letter->phone]} if /PHONE/;
				@email = @{[$letter->email]} if /EMAIL/;
				@url = @{[$letter->url]} if /URL/;
				@addr = @{[$letter->addr]} if /ADDR/;
			}
				if (defined @phone) {

				# remove duplicates
				my %phone ;
				undef %phone;
				@phone{@phone} = 0;
				@phone = (keys %phone);

				foreach (@phone) {

					 # trim 1- off front
					 s/^1-//;

					 # print file - date - phone
					 print "Phone : $file : $date_recvd : $_\n";
					 #print "$file : $_\n";
				}
			}
			if (defined @email) {

				# remove duplicates
				my %email ;
				undef %email;
				@email{@email} = 0;
				@email = (keys %email);

				foreach (@email) {

					 # print file - date - email
					 print "Email : $file : $date_recvd : $_\n";
				}
			}
				 if (@url) {
						# remove duplicates
				my %url ;
				undef %url;
				@url{@url} = 0;
			   @url = (keys %url);
				 foreach (@url) {
					  print "URL : $file : $_\n";
				 }
			 }
			 if (@addr) {
				 foreach (@addr) {
					  print "ADDR : $file : $date_recvd : $_\n";
				 }
			 }
		}
	  

=head1 AUTHOR
	
	Alan Jackson
	July 1998
	ajackson@icct.net

=cut

#
# Spamalyzer::Letter
#
#	Version 1.0 -  7/98 - Alan Jackson : ajackson at icct.net
#			 Copyright 1998 may be used and distributed under the
#			Gnu Copyleft.


package Spamalyzer::Letter;

use strict;
use vars qw( $VERSION );

use Carp;
use Date::Parse;
#use Mail::Address;

$VERSION = sprintf "%d.%02d", q$Revision: 1.0 $ =~ m#(\d+)\.(\d+)#;

sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my $self  = {};
    my $file;
    my $type;
	 my $isp;

    croak "usage: ${class}->new(FILENAME, TYPE, ISP)" unless @_ >= 3;

    $file  = shift;
	 $type = shift;
	 $isp = join(' ',@_);

		my @phone = '';
		my @email = '';
		my @url = '';
		my @addr = '';
    bless $self, $class;
    $self->source($file);
    $self->type($type);
    $self->isp($isp);
	 $self->{contacts} = [];
	 $self->contacts('NONE');
	 $self->{phone} = [];
	 $self->{email} = [];
	 $self->{url} = [];
	 $self->{addr} = [];

    if (defined $self->measure) {
        return $self;
    }
    else {
        return undef;
    }
}

sub source {
    my $self = shift;

    if (@_) {
	 	my $file = shift;
	 #print "---> $file\n";

		  open ART, $file or croak "$0: failed open [$file]: $!";
		  my $art = join '', <ART>;
		  close ART;
        $self->{source} = $art;
    }
    else {
        return $self->{source};
    }
}

sub type {
    my $self = shift;
    if (@_) { $self->{type} = shift; }
    else { return $self->{type}; }
}

sub isp {
    my $self = shift;
    if (@_) { $self->{isp} = join(' ',@_); }
    else { return $self->{isp}; }
}

sub measure {
    my $self = shift;
    my $art = $self->source;

    ################################
    # deal with headers
    ################################

    my $header;
    my %header;
	 my @phone;
	 my @email;
	 my @url;
	 my $addr;
	 my $goodhead = 1;

	 if ($self->type eq 'aol') {
	 	# some are forwards with a leading blank...
		  $art =~ s/\n ----/\n----/g;
		  $header = $art;
		  if ($art !~ /--- Headers ---/) { # internal AOL mail
				$goodhead = 0;

				$self->lastrelay('AOLINTERNAL');
				$self->lastrelayip('AOLINTERNAL');
				$self->prevrelay('NONE');
				$self->prevrelayip('NONE');
				$header =~ s/^.*Subj:/Subject:/;
				$header =~ s/\n\n.*$//;
		  }
				else {
			 # split aol mail
				#  I need to add in detection of internal AOL spam...
				$header =~ s/^([^\000]*)----------------------- Headers --------------------------------//;
				$art = $1;
				$header =~ s/\n\s*\n.*$//;
				$header =~ s/^\n*//;
				$header =~ s/Message-Id/Message-ID/g;
				$header =~ s/DATE/Date/g;
				$header =~ s/SUBJECT/Subject/g;
				$header =~ s/subject/Subject/g;
				$header =~ s/TO/To/g;
				# for a while, AOL indented the headers
				$header =~ s/\n ([A-Z][-\w]+:)/\n$1/g;
				#	anything after Subject: is bogus
				$header =~ s/(Subject:[^\n]*)\n[^\000]*/$1/;
		  }
	 }
	 else {
	 	# split normal mail

		  # CHOP!! right at the neck! :-)
		  unless ($art =~ s/^(.*?\n)\n//s) {
				warn "$0: no header in bogus article !";
				return undef;
		  }
		  $header = $1;
		  $header =~ s/Message-Id/Message-ID/g;
	 }

    ## fix continuation -- see perlfunc
    $header =~ s/\n\s+/ /g;
    %header = (dummy => split /^(.*?):/m, $header);
    chomp %header;  ## an abuse?

    # remove empty headers
    for (keys %header) {
	 	$header{$_} =~ s/^\s*//;
        delete $header{$_} unless $header{$_} =~ /\S/;
    }

    ## make sure all goes well with the headers
    unless ($self->_process_header(%header)) {

	 	print ">> header failer \n";
        return undef;
    }

	 # Process header for more difficult stuff

	 $self->header_extra($header) if $goodhead;
	 $self->date_recvd($header{'Date'}) if ! $goodhead;

    ################################
    # deal with the body 
    ################################
    my $body;

    $body = $art;         
    $art  = '';


	my @body = split(/\n/,$body);
	foreach (@body) {
		
		#####################################
		#	Look for US-style phone numbers
		#####################################

	 my @p = ();
	 if (/\D\d{3}\s*-\s*\d{4}[^-\d]/) { # line contains a phone number
		s/(\(|\)|-|\]|\[|\/)/ /g;
		@p = /(\D|^|\G)((1\s*)?(\d\d\d\s+)?\d{3}\s+\d{4})\D/g;
	 }
		
	 if (@p) {
		  $self->contacts('PHONE');
		  my $i;
		  for ($i=1;$i<$#p;$i+=4) {
				$p[$i] =~ s/^\s*//g;
				$p[$i] =~ s/\s*$//g;
				$p[$i] =~ s/ +/-/g;
				$p[$i] =~ s/^1(\d{3})/1-$1/;
				push @phone, $p[$i];
		  }
		  $self->phone(@phone);
	 }
		
		#####################################
		#	Look for e-mail addresses
		#####################################

	my @e = ();
	if (/@([-\w]+\.)+[\w]{2,3}(\W+|$)/) { # line contains an email address
		@e = /(^|\G|[^-\w])([-\w]+@([-\w]+\.)+[\w]{2,3})(\W+|$)/g;
	}
	if (@e) {
		 $self->contacts('EMAIL');
		 my $i;
		 for ($i=1;$i<$#e;$i+=4) {
			 $e[$i] =~ tr/A-Z/a-z/;
			 push @email, $e[$i];
		 }
		  $self->email(@email);
	}
		
		#####################################
		#	Look for URL's        
		#####################################

	my @u = ();
	@url = ();

	if (/(\.htm)|(www\.)|(http:)/i) { # line contains a url address
		chomp;
		s/http:\/\// /gi; # remove http stuff
		@u = /(^|\G|\W)((\w[-\w]*\.)+\w+(\/[~\w][-\w\.\/]*)*)([^-\w\.\/~]+|$)/g;
	}
	
	if (@u) {
		 $self->contacts('URL');
		 my $i;
		 for ($i=1;$i<$#u;$i+=5) {
			 $u[$i] =~ tr/A-Z/a-z/;
			 $u[$i] =~ s/\s*//g;
				# test completely numeric urls - is it an ip?	
			 if ($u[$i] !~ /\D/ && $u[$i] !~ /^(\d+\.){3}\d+$/) { next;}
			 	
			 push @url, $u[$i];
		 }
		  $self->url(@url);
	}
		
	}
		#####################################
		#	Look for street addresses
		#####################################

	my (%stateabbv, %statenames);
	my @linebuffer = (' ',' ',' ',' ',' '); # store last 5 lines

	 my @stateabbv  = qw ( AL AK AS AZ AR CA CO CT DE DC FM FL GA GU HI ID IL IN IA KS KY LA ME MH MD MA MI MN MS MO MT NE NV NH NJ NM NY NC ND MP OH OK OR PW PA PR RI SC SD TN TX UT VT VI VA WA WV WI WY );
	 for (@stateabbv) {$stateabbv{$_}=$_;}

	 %statenames = (
	 "ALABAMA" => "AL",
	 "ALASKA" => "AK",
	 "AMERICAN SAMOA" => "AS",
	 "ARIZONA" => "AZ",
	 "ARKANSAS" => "AR",
	 "CALIFORNIA" => "CA",
	 "COLORADO" => "CO",
	 "CONNECTICUT" => "CT",
	 "DELAWARE" => "DE",
	 "DISTRICT OF COLUMBIA" => "DC",
	 "FEDERATED STATES OF MICRONESIA" => "FM",
	 "FLORIDA" => "FL",
	 "GEORGIA" => "GA",
	 "GUAM" => "GU",
	 "HAWAII" => "HI",
	 "IDAHO" => "ID",
	 "ILLINOIS" => "IL",
	 "INDIANA" => "IN",
	 "IOWA" => "IA",
	 "KANSAS" => "KS",
	 "KENTUCKY" => "KY",
	 "LOUISIANA" => "LA",
	 "MAINE" => "ME",
	 "MARSHALL ISLANDS" => "MH",
	 "MARYLAND" => "MD",
	 "MASSACHUSETTS" => "MA",
	 "MICHIGAN" => "MI",
	 "MINNESOTA" => "MN",
	 "MISSISSIPPI" => "MS",
	 "MISSOURI" => "MO",
	 "MONTANA" => "MT",
	 "NEBRASKA" => "NE",
	 "NEVADA" => "NV",
	 "NEW HAMPSHIRE" => "NH",
	 "NEW JERSEY" => "NJ",
	 "NEW MEXICO" => "NM",
	 "NEW YORK" => "NY",
	 "NORTH CAROLINA" => "NC",
	 "NORTH DAKOTA" => "ND",
	 "NORTHERN MARIANA ISLANDS" => "MP",
	 "OHIO" => "OH",
	 "OKLAHOMA" => "OK",
	 "OREGON" => "OR",
	 "PALAU" => "PW",
	 "PENNSYLVANIA" => "PA",
	 "PUERTO RICO" => "PR",
	 "RHODE ISLAND" => "RI",
	 "SOUTH CAROLINA" => "SC",
	 "SOUTH DAKOTA" => "SD",
	 "TENNESSEE" => "TN",
	 "TEXAS" => "TX",
	 "UTAH" => "UT",
	 "VERMONT" => "VT",
	 "VIRGIN ISLANDS" => "VI",
	 "VIRGINIA" => "VA",
	 "WASHINGTON" => "WA",
	 "WEST VIRGINIA" => "WV",
	 "WISCONSIN" => "WI",
	 "WYOMING" => "WY", );

	my $j = 0;
	LINE: foreach (@body) {
			 chomp;
		 s/^\s*//;
		 unshift @linebuffer, $_; pop @linebuffer; # store line
		 $j++;

		 # search for a zipcode-like thing, and then work backwards from there.

		 if (/(^|\D)(\d{5})($|\D)/) { # line contains a potential zipcode
			 my $iline = 0;
			 my $zip = $2;
			 my $foo = $_; # keep line...
			 $addr = '';
			 # if it's a zip, then the next thing back ought to be a state...
			 $foo =~ s/,*([a-zA-Z ]+)\.*\s*$zip.*$//;
			 my $state = $1;
			 # if blank returned, back up a line.
			 if ($state =~ /^\s*$/) {
				 $iline = 1;
				 $foo = $linebuffer[$iline];
				 $foo =~ /,*([a-zA-Z ]+)\s*$/;
				 $state = $1;
			 }
			 my $is_state = 'no';
			 if (!defined $state || $state eq '') {next LINE;}
			 $state =~ s/^\s*//;
			 $state =~ s/\s*$//;
			 $state =~ s/\s+/ /;
			 $state =~ tr/a-z/A-Z/;
			 #	Is it really a state?
			 if (length($state) == 2) {
				 if (defined $stateabbv{$state}) {$is_state = 'yes';}
			 }
			 else {
				 if (defined $statenames{$state}) {$is_state = 'yes';}
				 elsif ($state =~ / /) { # remove leading stuff
					 while ($state =~ / /) {
						 $state =~ s/^\S+\s+//;
						 if (defined $statenames{$state}) {$is_state = 'yes';}
						 if (defined $stateabbv{$state}) {$is_state = 'yes';}
					 }
				 }
				 else {$is_state = 'no';}
			 }
			 if ($is_state eq 'no') {next LINE;} # not an address after all

			  $self->contacts('ADDR'); # We have an address!

			 # big assumption follows - city preceeds state and if there is
			 # no comma or digit in front of state, then it *all* is city
			 # and the preceeding 3 lines are likely address. Otherwise it
			 # is a one line address... see I said it was a big assumption!

			 # nuke state and everything to the right

			 $linebuffer[$iline] =~ s/[, ]$state[, ].*$//i;
			 $linebuffer[$iline] =~ s/[^a-zA-Z]*$//;

			 # is there a comma in what is left? For that matter, is anything left?

			$foo = $linebuffer[$iline];
			#	Test for a street abbrev or PO box
			my $street = 'no';
			if ($foo =~ / (st|ave|blvd|pl|dr|ct|rd|way|road|drive|cir)\.* /i || $foo =~ / box\.*/i) {$street = 'yes';}
			 if (($foo !~ /,/ || $foo =~ /^\s*$/) && $street eq 'no') {# back up 3 lines
				 $addr = '';
				 my $i;
				 for ($i=$iline+3;$i>=$iline;$i--) {
					 $linebuffer[$i] =~ s/<[^<>]+>//g;
					 $addr = $addr . "," . $linebuffer[$i];
				 }
				 $addr =~ s/^.*://;
				 	# replace state name with standard abbreviation
				 if (length($state) != 2) {$state = $statenames{$state};}
				 $addr .= " : " . $state . " : " . $zip;
			 }
			 else {
				 $linebuffer[$iline] =~ s/^.*://;
				 	# replace state name with standard abbreviation
				 if (length($state) != 2) {$state = $statenames{$state};}
				 $addr = $linebuffer[$iline] . " : " . $state . " : " . $zip;
			 }
			 $addr =~ s/^[^[A-Za-z]*//;
			 $addr =~ s/,+/,/g;
			 $addr =~ s/\s+/ /g;
			 $addr =~ s/-+/-/g;
		    $self->addr($addr);
			 
		 }
	 }

    1;
}

sub date {
    my $self = shift;

    if (@_) {
        my $date;
        ###$date = str2time shift;
        $date = shift;
        $self->{date} = $date;

    }
    else {
        return $self->{date};
    }

    1;
}

sub contacts {
    my $self = shift;
    if (@_) { unshift @{$self->{contacts}}, @_; }
    else { return @{$self->{contacts}}; }
}

sub subject {
    my $self = shift;
    if (@_) { $self->{subject} = shift; }
    else { return $self->{subject}; }
}

sub phone {
    my $self = shift;
    if (@_) { push @{$self->{phone}}, @_; }
    else { return @{$self->{phone}}; }
}

sub email {
    my $self = shift;
    if (@_) { push @{$self->{email}}, @_; }
    else { return @{$self->{email}}; }
}

sub url {
    my $self = shift;
    if (@_) { push @{$self->{url}}, @_; }
    else { return @{$self->{url}}; }
}

sub addr {
    my $self = shift;
    if (@_) { push @{$self->{addr}}, @_; }
    else { return @{$self->{addr}}; }
}

sub from {
    my $self = shift;
    if (@_) { $self->{from} = shift; }
    else { return $self->{from}; }
}

sub message_id {
    my $self = shift;
    if (@_) { $self->{message_id} = shift; }
    else { return $self->{message_id}; }
}

sub reply_to {
    my $self = shift;
    if (@_) { $self->{reply_to} = shift;}
    else { return $self->{reply_to}; }
}

sub return_path {
    my $self = shift;
    if (@_) { $self->{return_path} = shift;}
    else { return $self->{return_path}; }
}

sub lastrelay {
    my $self = shift;
    if (@_) { $self->{lastrelay} = shift;}
    else { return $self->{lastrelay}; }
}

sub lastrelayip {
    my $self = shift;
    if (@_) { $self->{lastrelayip} = shift;} 
    else { return $self->{lastrelayip}; }
}

sub date_recvd {
    my $self = shift;
	 if (@_) { $self->{date_recvd} = shift;}
	 else { return $self->{date_recvd}; }
}

sub prevrelay {
    my $self = shift;
	 if (@_) { $self->{prevrelay} = join(' ',@_);}
	 else { return $self->{prevrelay}; }
}

sub prevrelayip {
    my $self = shift;
	 if (@_) { $self->{prevrelayip} = shift;}
	 else { return $self->{prevrelayip}; }
}

sub _process_header {
    my $self   = shift;
    my %header = @_;
    my $file   = $self->source;

    ################################
    # Date:
    ################################
    if (! exists $header{Date}) {
	 	$header{Date} = 'NONE';
    }
    $self->date($header{Date});

    ################################
    # From:
    ################################
    if (! exists $header{From}) {
	 	$header{From} = 'NONE';
    }
    $self->from($header{From});

    ################################
    # Subject:
    ################################
    if (! exists $header{Subject}) {
	 	$header{Subject} = 'NONE';
    }
    $self->subject($header{Subject});

    ################################
    # Message-ID:
    ################################
    if (! exists $header{'Message-ID'}) {
	 	$header{'Message-ID'} = 'NONE';
    }
    $self->message_id($header{'Message-ID'});

    ################################
    # Reply-to:  
    ################################
    if (! exists $header{'Reply-to'}) {
	 	$header{'Reply-to'} = 'NONE';
    }
    $self->reply_to($header{'Reply-to'});

    ################################
    # Return_Path:
    ################################
    if (! exists $header{'Return-path'}) {
	 	$header{'Return-path'} = 'NONE';
    }
    $self->return_path($header{'Return-path'});

    1;
}

sub header_extra {
    my $self   = shift;
    my $header = shift;
	 my $isp = $self->isp;
	 my $type = $self->type;


    ################################
    # Last Relay 
    ################################

	 #	Use the ISP specific pattern to pick out the correct header record

	my @header = split(/\n/,$header);
	my $i = 0;
	my ($last_relay,$last_relay_ip) ;

	if ($type eq 'normal') {
		 for (@header) {
			 if (/$isp/) {
			 	if (/unknown/ && /HELO/) {
					  ($last_relay) = ($_ =~ /HELO (\S+)\)/);
				}
				else {
					  ($last_relay) = ($_ =~ /^Received: from ([^\s(]+)[ (]/);
				}
				 ($last_relay_ip) = ($_ =~ /\(([\d.]+)\)/);
				 last;
			 }
			 $i++;
		 }
	 }
	 elsif ($type eq 'aol') { # aol headers - look for "from !aol by aol"
	   my $aol_internal = 'yes';
		for (;$i<$#header;$i++) {
		 	$_ = $header[$i];
		 	next if !/^Received:/;
			$aol_internal = 'no';
			my $fromcpu;
			if (/\(\[/) {($fromcpu) = /from\s+(\S+)\s*\(/;}
			else {($fromcpu) = /from\s+[^(]+\((\S+)\s*[[( ]/;}
			my ($bycpu) = /by\s+(\S+)/;
			next if (defined $fromcpu && $fromcpu =~ /aol\.com/);
			if (defined $fromcpu) {$last_relay = $fromcpu;}
			else {$last_relay = 'NONE';}
			($last_relay_ip) = /\[([\d.]+)\]/;
			last;
		}
	 	#		If it began within aol, we have to handle special
	 	
		if ($aol_internal eq 'yes') {
			$last_relay_ip = 'NONE';
			$last_relay = 'AOL';
		}
		if (! defined $last_relay_ip) {
			$i = 0;
			if ($header[$i] !~ /^Received/) {$i++;}
			if ($header[$i] !~ /^Received/) {$i++;}
			$_ = $header[$i];
			($last_relay) = /from\s+\S+\s+\((\S+)\s*\[/;
			($last_relay_ip) = /\[([\d.]+)\]/;
		}
	 }


    $self->lastrelay($last_relay);
    $self->lastrelayip($last_relay_ip);

    ################################
    # Date_Recvd 
    ################################
    
	 my $date_recvd ;
	 ($date_recvd) = ($header[$i] =~ /;(.*)$/);
	 if ($last_relay eq 'AOL') {$date_recvd = $self->date;}

	  if (! defined $last_relay) {my $file = $self->source;print STDERR "Header E
rror in $file\n";}
	 
	 $self->date_recvd($date_recvd);

    ################################
    # Prev Relay 
    ################################
    
	 my ($prev_relay,$prev_relay_ip) ;
	 my $prevrelayip = 'null';
	 my ($target) = ($last_relay =~ /([-\w\d]+\.?\w*$)/);
	 $target =~ s/\./\\./g;
	 $i++;
	 for (;$i<$#header;$i++) {
		 next if $header[$i] !~ /^Received:/;
		 next if $header[$i] =~ /localhost/; # internal handoff
		 last if ($header[$i] =~ /$target/); # found next relay
	 }
	 $prev_relay = $header[$i];
	 if ($last_relay eq 'AOL') {$prev_relay = 'NONE';}
	 if ($prev_relay !~ /^Received/) {$prev_relay = 'NONE';} # trail ends
	 elsif ($prev_relay !~ /from/) {$prev_relay = 'BROKEN';} # doesn't id source
	 $prev_relay =~ s/;.*$//; # nuke date field
	 $prev_relay =~ s/^Received://; # nuke Received: 
	 # format is either from system by system, or by system from system
	 $prev_relay =~ s/^.*from//;
	 $prev_relay =~ s/ by .*$//;

	 #		Now try to extract the ip address and machine name, if present

	 ($prev_relay_ip) = ($prev_relay =~ /(\d+\.\d+\.\d+\.\d+)/);
	 if (defined $prev_relay_ip) { # remove ip address and everything to right
	 	my $ip = $prev_relay_ip ;
	 	$ip =~ s/\./\\./g;
	 	$prev_relay =~ s/$ip.*$//;
	 }
	 else { $prev_relay_ip='NONE';}

	 $prev_relay =~ s/\(HELO .*$//; # Are the HELO's at this stage usually bogus?
	 # clean up parens and brackets
	 $prev_relay =~ s/[\[\]\(\)]/ /g;
	 if ($prev_relay =~ /^\s*$/) {$prev_relay = 'NONAME';}
	 # take rightmost field as machine name
	 $prev_relay = (split(/\s+/,$prev_relay))[-1];
	 
	 $self->prevrelay($prev_relay);
	 $self->prevrelayip($prev_relay_ip);

    ################################
    # Words        
    ################################

    1;
}

1;

